/**
 * @file oainkubator.cpp
 */

/*
 * Copyright (C) Gemfony scientific UG (haftungsbeschraenkt)
 *
 * See the AUTHORS file in the top-level directory for a list of authors.
 *
 * Contact: contact [at] gemfony (dot) eu
 *
 * This file is part of the Geneva library collection.
 *
 * Geneva was developed with kind support from Karlsruhe Institute of
 * Technology (KIT) and Steinbuch Centre for Computing (SCC). Further
 * information about KIT and SCC can be found at http://www.kit.edu/english
 * and http://scc.kit.edu .
 *
 * Geneva is free software: you can redistribute and/or modify it under
 * the terms of version 3 of the GNU Affero General Public License
 * as published by the Free Software Foundation.
 *
 * Geneva is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with the Geneva library. If not, see <http://www.gnu.org/licenses/>.
 *
 * For further information on Gemfony scientific and Geneva, visit
 * http://www.gemfony.eu .
 */

/*
 * This code is meant as an incubator that facilitates "playing" with new
 * optimization algorithms. It uses some of the conventions common in the
 * Geneva optimization library (see http://launchpad.net/geneva), and thus
 * may be useful if you wish to build new algorithms to be integrated into
 * Geneva. It currently only supports double-type parameters. If you want
 * other parameter types, you can "emulate" them -- 0. is false and 1. is
 * true for boolean types, 0.,1.,2. ... would represent integer types.
 *
 * Some code in this file was contributed by Lisa Schaetzle.
 */

#include <iostream>
#include <vector>
#include <cmath>
#include <limits>

/******************************************************************************/
/**
 * This enum denotes the possible demo function types
 */
enum solverFunction {
   PARABOLA=0
   , NOISYPARABOLA=1
   , ROSENBROCK=2
   , ACKLEY=3
   , RASTRIGIN=4
   , SCHWEFEL=5
   , SALOMON=6
};

/******************************************************************************/
/**
 * Calculation of the fitness of a given solution
 */
class solver {
public:
   solver(const solverFunction& f)
      :f_(f)
   { /* nothing */ }

   // The copy constructor
   solver(const solver& cp)
      :f_(cp.f_)
   { /* nothing */ }

   // virtual destructor, so we can later derive classes from this one
   virtual ~solver() { /* nothing */ }

   // This is just syntactic sugar ...
   double operator()(const std::vector<double>& vec) const {
      return fitnessCalculation(vec);
   }

   // Trigger the actual fitness calculation
   double fitnessCalculation(const std::vector<double> vec) const {
      double result = 0.;

      switch(f_) {
         case PARABOLA:
         result = fitnessParabola(vec);
         break;

         case NOISYPARABOLA:
         result = fitnessNoisyParabola(vec);
         break;

         case ROSENBROCK:
         result = fitnessRosenbrock(vec);
         break;

         case ACKLEY:
         result = fitnessAckley(vec);
         break;

         case RASTRIGIN:
         result = fitnessRastrigin(vec);
         break;

         case SCHWEFEL:
         result = fitnessSchwefel(vec);
         break;

         case SALOMON:
         result = fitnessSalomon(vec);
         break;
      }

      return result;
   }

protected:
   double fitnessParabola(const std::vector<double>& vec) const {
      double result = 0;

      for(std::size_t i=0; i<vec.size(); i++) {
         result += vec.at(i)*vec.at(i);
      }

      return result;
   }

   double fitnessNoisyParabola(const std::vector<double>& vec) const {
      double result = 0;

      // Calculate result --> You can get all solver function definitions from here:
      // http://bazaar.launchpad.net/~gemfony/geneva/trunk/view/head:/src/geneva-individuals/GFunctionIndividual.cpp
      // For the purpose of this demo, you can copy them verbatim (look at line 871 ff.). The macro "GSQUARED(x)" can
      // be replaced with pow(x,2.)

      double xsquared = 0.;
      for(std::size_t i=0; i<vec.size(); i++){
            xsquared += pow(vec.at(i),2.);
      }
      result = (cos(xsquared) + 2.) * xsquared;
      return result;
   }

   double fitnessRosenbrock(const std::vector<double>& vec) const {
      double result = 0;

      for(std::size_t i=0; i<(vec.size()-1); i++) {
            result += 100.*pow((pow(vec.at(i),2.) - vec.at(i+1)),2.) + pow(1.-vec.at(i),2.);
      }
      return result;
   }

   double fitnessAckley(const std::vector<double>& vec) const {
      double result = 0;
      for(std::size_t i=0; i<(vec.size()-1); i++) {
         result += (exp(-0.2)*sqrt(pow(vec.at(i),2.) + pow(vec.at(i+1),2.)) + 3.*(cos(2.*vec.at(i)) + sin(2.*vec.at(i+1))));
      }
      return result;
   }

   double fitnessRastrigin(const std::vector<double>& vec) const {
      double result = 10*double(vec.size());

      for(std::size_t i=0; i<vec.size(); i++) {
            result += (pow(vec.at(i),2.) - 10.*cos(2*M_PI*vec.at(i)));
      }

      return result;
   }

   double fitnessSchwefel(const std::vector<double>& vec) const {
      double result = 0;

      for(std::size_t i=0; i<vec.size(); i++) {
         result += -vec.at(i)*sin(sqrt(fabs(vec.at(i))));
      }

      result /= vec.size();

      return result;
   }

   double fitnessSalomon(const std::vector<double>& vec) const {
      double result = 0;

      double sum_root = 0.;
      for(std::size_t i=0; i<vec.size(); i++) {
            sum_root += pow(vec.at(i),2.);
      }

      sum_root=sqrt(sum_root);
      result = -cos(2*M_PI*sum_root) + 0.1*sum_root + 1;

      return result;
   }

   solverFunction f_; ///< The chosen solver

private:
   // The default constructor: intentionally private and undefined, as we want to enforce usage of solver(const solverFunction& f)
   solver();
};

/******************************************************************************/
/**
 * The optimizer base class. Note that this function assumes that your
 * optimization algorithm MINIMIZES only.
 */
class optimizerBase {
public:
   // The default constructor
   optimizerBase(
      const std::vector<double>& startValues
      , const solver& s
      , const std::size_t& maxIterations
   )
      : bestEvaluation_(std::numeric_limits<double>::max())
		, bestParameters_(startValues)
      , solver_(s)
      , maxIterations_(maxIterations)
      , currentIteration_(0)
   { /* nothing */ }

   // The copy constructor
   optimizerBase(const optimizerBase& cp)
      : bestEvaluation_(cp.bestEvaluation_)
	  	, bestParameters_(cp.bestParameters_)
      , solver_(cp.solver_)
      , maxIterations_(cp.maxIterations_)
      , currentIteration_(0)
   { /* nothing */ }

   // virtual destructor
   virtual ~optimizerBase() { /* nothing */ }

	// Retrieve the current iteration
	std::size_t getCurrentIteration() {
		return currentIteration_;
	}

	// Retrieve the best result found so far
	double getBestResult() {
		return bestEvaluation_;
	}

	// Retrieve the best parameters  found so far
	std::vector<double> getBestParameters() {
		return bestParameters_;
	}

   // The external optimizer interface.
   // Will reset the currentIteration_ variable when called
	// Returns a vector of best solutions.
   std::vector<double> optimize() {
		currentIteration_ = 0;
		double currentEvaluation = std::numeric_limits<double>::max();
		std::vector<double> currentBestParameters(bestParameters_.size());

		// Initialization code
		init();

		// Optimize, until a halt criterion is reached
		do {
			// This is where the actual work is done
			currentEvaluation = cycleLogic(currentBestParameters);

			// Update the best results
			if(currentEvaluation <= bestEvaluation_) {
				bestEvaluation_ = currentEvaluation;
				bestParameters_ = currentBestParameters;
			}

			// Emit progress information
			std::cout << this->getCurrentIteration() << ": " << this->getBestResult() << std::endl;
		} while(!this->halt());

		// Finalization code
		finalize();

		// Let the audience know
		return this->getBestParameters();
   }

protected:
   // Overload this function in derived classes, if initialization work is required
   virtual void init() { /* nothing */ };
   // Overload this function in derived classes, if initialization work is required
   virtual void finalize() { /* nothing */ };
	// Overload this function in derived classes -- it holds the logic to be executed in each iteration
	virtual double cycleLogic(std::vector<double>&) = 0;

   // Return "true", if currentIteration_ becomes larger than maxIterations_, otherwise "false"
   bool halt() { return (currentIteration_++ >= maxIterations_); }

	double bestEvaluation_; ///< Holds the best value found so far
   std::vector<double> bestParameters_; ///< Holds the currently best parameter set

   solver solver_; ///< Holds the solver object

private:
   std::size_t maxIterations_; ///< Holds the maximum number of optimization cycles -- this will serve as our stop criterion for the optimization
   std::size_t currentIteration_; ///< The current iteration being processed

   // The default constructor: intentionally private and undefined, as we want to enforce usage of optimizerBase(const std::size_t&)
   optimizerBase();
};

/******************************************************************************/
/**
 * A place holder for optimization algorithms to be tried out
 */
class optimizerPlaceHolder
   : public optimizerBase
{
public:
   optimizerPlaceHolder(
      const std::vector<double>& startValues
      , const solver& s
      , const std::size_t& maxIterations
   )
      : optimizerBase(startValues, s, maxIterations) // Initialize our parent class
   { /* nothing */ }

protected:
	// The optimization logic. We do nothing in this dummy optimizer. You may implement
	// your own optimization code here.
	virtual double cycleLogic(std::vector<double>& bestParameters) {
		bestParameters = bestParameters_;
		return bestEvaluation_;
	}

private:
   // The default constructor: intentionally private and undefined, as we want to enforce usage of optimizerBase(const std::size_t&)
   optimizerPlaceHolder();
};

/******************************************************************************/
/**
 * Helper function to output results and keep "main()" clean
 */
void print(const std::vector<double>& vec, const std::string& envelope) {
   std::cout << envelope << std::endl;
   std::vector<double>::const_iterator cit;
   for(cit=vec.begin(); cit!=vec.end(); ++cit) {
      std::cout << *cit << " ";
   }
   std::cout << std::endl;
}

/******************************************************************************/
/**
 * Some more syntactic sugar
 */
const int SUCCESS = 0;
const int FAILURE = 1;
const std::size_t MAXITERATIONS = 10;

/******************************************************************************/
/**
 * The main function (who would have guessed ...)
 */
int main() {
   std::vector<double> startValues;
   startValues.push_back(1.);
   startValues.push_back(2.);
   startValues.push_back(3.);
   startValues.push_back(4.);

   std::vector<double> bestResults;

   // Set up our solver
   solver s(PARABOLA);
   //-----------------------------------------------------
   // We start to optimize with the conjugate gradient
   optimizerPlaceHolder cg(startValues, s, MAXITERATIONS);

   // Run the actual optimization
   bestResults = cg.optimize();

   // Output the result
   print(bestResults, "DummyOA:");

   //-----------------------------------------------------
   // We are done -- let the audience know

   std::cout << "done ..." << std::endl;
   return SUCCESS;
}
